--------------------------------------------------------------------------------
-- Demolition Derby AI Lua Script
-- By Scepheo
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Scenario config (read from json)
--------------------------------------------------------------------------------

-- Name of each opponent
local opponentName


--------------------------------------------------------------------------------
-- AI config
--------------------------------------------------------------------------------

-- Minimum speed a vehicle has to have before counting as standing still (m/s)
local minimumSpeed = 0.5 -- m/s

-- Seconds before a player is timed out
local maxTimeout = 31

-- Number of opponents
local opponentCount = 3

-- Number of AI target nodes to place in either direction
local xCount = 30
local yCount = 20

-- How many "outer" nodes there are: nodes the AI shouldn't drive to
local nodeOffset = 5

-- The radius of the generated nodes
local aiNodeRadius = 3

-- Minimum distance the AI should be from its target before attacking
local aiMinTargetDistance = 10

-- How close the AI has to get to its target vehicle to count as a "hit"
local aiMaxTargetDistance = 5

-- How close the AI has to get to its target node to count as "reached"
local aiMaxNodeDistance = 5

-- The maximum angle an AI is allowed to have to its target, before disengaging
local aiMaxTargetAngle = 110

-- the distance an AI aims past its target, to ensure good speed at contact
local aiOveraimDistance = 10

-- AI aggression setting
local aiAggression = 1

-- How far an AI targets when "backing out" of a crash
local aiBackoutDistance = 5

--------------------------------------------------------------------------------
-- Scenario data
--------------------------------------------------------------------------------
local player = nil
local playerTimeout = 0

--------------------------------------------------------------------------------
-- AI data
--------------------------------------------------------------------------------
local opponent = {}
local opponentState = {}
local opponentCrashed = {}
local opponentTimeout = {}

--------------------------------------------------------------------------------
-- Node functions
--------------------------------------------------------------------------------

-- Contains all nodes generated by the scenario
local nodes = nil

-- Creats a grid of nodes on the map
local function createNodes()
	nodes = {}

	local sw = scenetree.findObject('wp_sw'):getPosition()
	local ne = scenetree.findObject('wp_ne'):getPosition()
	
	local width = ne.x - sw.x
	local height = ne.y - sw.y
	
	local z = ne.z
	
	-- Create all nodes
	for i = 0, xCount do
	
		local dx = width * i / xCount
		local x = sw.x + dx
		
		for j = 0, yCount do
			
			local dy = height * j / yCount
			local y = sw.y + dy
			
			local nodeName = i .. 'x' .. j
			local pos = Point3F(x, y, z)
			
			local node =
			{
				pos = Point3F(x, y, z),
				radius = aiNodeRadius,
				segTags = {},
				links = {},
				manual = 1,
			}
			
			nodes[nodeName] = node
		end
	end
	
	print('Done creating nodes')
	
	-- Link all nodes
	for i = 0, xCount do
		for j = 0, yCount do
			local nodeName = i .. 'x' .. j
			local node = nodes[nodeName]
			
			for di = -1, 1 do
				for dj = -1, 1 do
					local neighbourName = tostring(i + di) .. 'x' .. tostring(j + dj)
					local neighbour = nodes[neighbourName]
					
					if neighbour then
						node.links[neighbourName] = 1
					end
				end
			end
		end
	end
	
	print('Done linking nodes')
	
	-- Add all nodes
	local mapInstance = map.getMap()
	
	for thisName, thisNode in pairs(nodes) do
		if not mapInstance.nodes[thisName] then
			mapInstance.nodes[thisName] = thisNode
		end
	end
	
	print('Done adding nodes')
end

-- Returns the name of the node closest to the given position.
local function closestNode(position)
	local bestNode = nil
	local bestDist = math.huge
	
	for name, node in pairs(nodes) do
		local dist = math.abs((position - node.pos):len())
		
		if dist < bestDist then
			bestNode = name
			bestDist = dist
		end
	end
	
	return bestNode
end

-- Returns a random node for the AI to drive to.
local function getRandomNode()

	-- We don't want the outer nodes, those are only for  targeting past cars
	local i = math.random(nodeOffset, xCount - nodeOffset)
	local j = math.random(nodeOffset, yCount - nodeOffset)
	
	return i .. 'x' .. j
end

-- Returns a list of all possible vehicles to target for the given AI, in random
-- order.
local function getTargetListFor(ai)
	local targets = {}
	
	for i = 1, opponentCount do
		if opponent[i] ~= ai and not opponentCrashed[i] then
			table.insert(targets, opponent[i])
		end
	end
	
	table.insert(targets, player)
	
	for i = 2, #targets do
		local j = math.random(1, i)
		
		if i ~= j then
			targets[i], targets[j] = targets[j], targets[i]
		end
	end
	
	return targets
end

--------------------------------------------------------------------------------
-- Vector math
--------------------------------------------------------------------------------

-- Gets the angle (in degrees) between the two vectors, projected on the
-- xy-plane.
local function getAngle(v, w)
	local vl = math.sqrt(v.x * v.x + v.y + v.y)
	local wl = math.sqrt(w.x * w.x + w.y + w.y)
	
	local dot = (v.x / vl) * (w.x / wl) + (v.y / vl) * (w.y / wl)
	local angle = math.deg(math.abs(math.acos(dot)))
	
	return angle
end

-- Divide vector by scalar
local function vecDiv(vector, value)
	return Point3F(vector.x / value, vector.y / value, vector.z / value)
end

-- Multiple vector by scalar
local function vecMul(vector, value)
	return Point3F(vector.x * value, vector.y * value, vector.z * value)
end

--------------------------------------------------------------------------------
-- AI code
--------------------------------------------------------------------------------

local function handleAI(i)
	-- Shortcuts
	local ai = opponent[i]
	local data = map.objects['opponent' .. i]
	local state = opponentState[i]
	
	local aiPosition = ai:getPosition()
	
	-- Try going to a random node after a long time of standing still
	if opponentTimeout[i] > 6 then
		if opponentTimeout[i] % 2 == 0 then
			state.targetNode = getRandomNode()
			state.targetCar = nil
			state.targetPosition = nodes[state.targetNode].pos
			
			ai:queueLuaCommand('ai.setTarget("' .. state.targetNode.. '")')
		end
		
		return
	
	-- Try going backwards after a short time of standing still
	elseif opponentTimeout[i] > 2 then
		targetPosition = aiPosition - vecMul(data.dirVec, aiBackoutDistance)
			
		state.targetNode = closestNode(targetPosition)
		state.targetPosition = targetPosition
		state.targetCar = nil
		
		ai:queueLuaCommand('ai.setTarget("' .. state.targetNode.. '")')
		
		return
	end
	
	-- Logic when we have a target
	if state.targetPosition then
		
		-- We're targeting a car
		if state.targetCar then
			local targetPosition = state.targetCar:getPosition()
			local difference = targetPosition - aiPosition
			local targetDistance = math.abs(difference:len())
			
			if targetDistance < aiMaxTargetDistance then
			
				print(opponentName[i] .. ' reached car')
				
				-- Back away from our target
				targetPosition = aiPosition - vecMul(data.dirVec, aiBackoutDistance)
				state.targetNode = closestNode(targetPosition)
				state.targetPosition = targetPosition
				state.targetCar = nil
				ai:queueLuaCommand('ai.setTarget("' .. state.targetNode.. '")')
				
			else
				
				-- See if our target is still somewhat in front of us.
				local targetAngle = getAngle(difference, data.dirVec)
								
				if targetAngle > aiMaxTargetAngle then
					print(opponentName[i] .. ' disengaging ' .. state.targetCar.name)
					state.targetNode = nil
					state.targetCar = nil
					state.targetPosition = nil
					return
				end
				
				difference = vecDiv(difference, difference:len())
				difference = vecMul(difference, aiOveraimDistance)
				targetPosition = targetPosition + difference
			
				state.targetNode = closestNode(targetPosition)
				state.targetPosition = targetPosition
				
				ai:queueLuaCommand('ai.setTarget("' .. state.targetNode.. '")')
			end
			
		-- We're targeting a node
		else
			-- See if we've reached our target
			local targetDistance  = math.abs((aiPosition - state.targetPosition):len())
			
			if targetDistance < aiMaxNodeDistance then
				state.targetNode = nil
				state.targetCar = nil
				state.targetPosition = nil
				
				print(opponentName[i] .. ' reached target')
			end
		end
		
	-- Logic for when we don't have a target
	else
		-- Try targets in random order
		local targetList = getTargetListFor(ai)
		local foundTarget = false
		
		for i, potentialTarget in pairs(targetList) do
		
			-- Determine our distance to the target
			local targetPosition = potentialTarget:getPosition()
			local aiPosition = ai.getPosition()
			local distance = math.abs((aiPosition - targetPosition):len())
			
			-- Determine our angle to the target
			local targetAngle = getAngle(targetPosition - aiPosition, data.dirVec)
		
			-- Only target after some minimum distance is achieved
			if distance > aiMinTargetDistance and targetAngle < aiMaxTargetAngle then
				local difference = targetPosition - aiPosition
				difference = vecDiv(difference, difference:len())
				difference = vecMul(difference, aiOveraimDistance)
				targetPosition = targetPosition + difference
			
				state.targetNode = closestNode(targetPosition)
				state.targetCar = potentialTarget
				state.targetPosition = targetPosition
				
				ai:queueLuaCommand('ai.setTarget("' .. state.targetNode.. '")')
				
				print(opponentName[i] .. ' attacking ' .. state.targetCar.name .. ' at ' .. state.targetNode)
				
				foundTarget = true
				break
			end
			
		end
		
		if not foundTarget then
			state.targetNode = getRandomNode()
			state.targetCar = nil
			state.targetPosition = nodes[state.targetNode].pos
			
			ai:queueLuaCommand('ai.setTarget("' .. state.targetNode.. '")')
			
			print(opponentName[i] .. ' driving to random node ' .. state.targetNode)
		end
	end
end

--------------------------------------------------------------------------------
-- Scenario code
--------------------------------------------------------------------------------

local M = {}

local running = false

local function reset()
    playerTimeout = 0
	
	for i = 1, opponentCount do
		opponentState[i] =
		{
			hasNode = false
		}
		opponentCrashed[i] = false
		opponentTimeout[i] = 0
	end
	
	running = false
end

local function onScenarioChange(scenario)
	math.randomseed(os.time())
	
	if scenario.state == 'pre-start' then
		createNodes()
		opponentName = scenario.demoNames
	end
end

local function onRaceStart()
    reset()
    running = true
    
    player = scenetree.findObject('scenario_player0')
    player:queueLuaCommand('mapmgr.enableTracking("player")')
	
	for i = 1, opponentCount do
		opponent[i] = scenetree.findObject('opponent' .. i)
		opponent[i]:queueLuaCommand('ai.setState({mode="manual"})')
		opponent[i]:queueLuaCommand('mapmgr.enableTracking("opponent' .. i.. '")')
		opponent[i]:queueLuaCommand('ai.setAggression(' .. aiAggression .. ')')
	end
end

local function fail(reason)
    scenarios.finish({failed = reason})
    reset()
end

local function success(reason)     
    scenarios.finish({msg = reason})
    reset()
end

local function onRaceTick(raceTickTime)  
    if not running then
        return
    end
    
    local playerData = map.objects.player
    
    if not playerData then
        return
    end
    
    local playerSpeed = playerData.vel:length()
	
	if playerSpeed < minimumSpeed then 
        playerTimeout = playerTimeout + raceTickTime
		
		if playerTimeout % 1 == 0 then
            local countdown = maxTimeout - math.modf(playerTimeout)
			
            if countdown > 0 then
                guihooks.trigger('ScenarioFlashMessage', {{countdown .. '...', 0.75}} )
            end
        end
		
    else
        playerTimeout = 0
    end
	
    if playerTimeout >= maxTimeout then
        fail("You were timed out!")
    end
	
	-- Handle opponents
	local opponentRemaining = false
	
	for i = 1, opponentCount do
		-- Check if he's still going
		if not opponentCrashed[i] then
			opponentRemaining = true
		end
	
		local opponentData = map.objects['opponent' .. i]
		
		if opponentData and not opponentCrashed[i] then
			-- Fetch this opponent's speed
			local speed = opponentData.vel:length()
			
			-- Count down
			if speed < minimumSpeed then
				opponentTimeout[i] = opponentTimeout[i] + raceTickTime
			else
				opponentTimeout[i] = 0
			end
			
			-- Check if he's timed out
			if opponentTimeout[i] >= maxTimeout then
				opponentCrashed[i] = true
				opponent[i]:queueLuaCommand('ai.setState({mode="stop"})')
				
				local message = opponentName[i] .. ' timed out!'
				guihooks.trigger('ScenarioFlashMessage', {{ message, 2.5 }})
			else
				handleAI(i)
			end
		end
	end
	
	if not opponentRemaining then
		success("You're the last man standing!")
	end
end

M.onRaceStart = onRaceStart
M.onRaceTick = onRaceTick
M.onScenarioChange = onScenarioChange

return M